class ExperimentDriver(object):
    def __init__(self, seed_job=None, max_running_jobs=100):
        # The list of possible status shall cover all the possible situations in which a job can be found.
        # PENDING: The job has been created and ready to be launched.
        # RUNNING: The job is RUNNING.
        # SUCCESS: A job that has been completed with successful results and asserted.
        # FAILED: A job that has been completed with unsuccessful results and that can not be
        # split ( Already was or a single variable set).
        # SUSPENDED: A job that has been completed with unsuccessful results that has been split and depends on the
        # results of descendant jobs.
        self.loop = 0
        self.pending = []
        self.running = []
        self.success = []
        self.failed = []
        self.suspended = []
        self.disinherited = []
        self.__dict = {"PENDING": self.pending,
                       "RUNNING": self.running,
                       "SUCCESS": self.success,
                       "FAILED": self.failed,
                       "SUSPENDED": self.suspended,
                       "DISINHERITED": self.disinherited,
                       }
        self.seed_job = seed_job
        self.max_running_jobs = max_running_jobs
        if self.seed_job is not None:
            self.pending.append(self.seed_job)
            self.checkpoint_name = "checkpoint_%i.pkl" % self.__hash__()
        else:
            self.checkpoint_name = "checkpoint.pkl"

    def sort(self, lower_first=True):
        if lower_first:
            self.pending.sort(key=lambda _x: _x.level)
        else:
            self.pending.sort(key=lambda _x: - _x.level)

    def all(self):
        return self.running + self.pending + self.success + self.failed + self.suspended

    def __getitem__(self, item):
        return self.__dict[item]

    def step(self):
        """
        The function will loop through the jobs in the queue and will execute the driver.
        :return: Number of jobs that changed status
        """
        changes = 0
        for job in self.all():
            if job.driver(queue=self, max_running_jobs=self.max_running_jobs): changes += 1
        self.loop += changes
        return changes

    def print_status(self):
        for key in self.__dict.keys():
            print("\t%18s: %4i" % (key, len(self[key])))

    def run(self, condition):
        import time

        loops = 0
        change_counter = 0
        wait = False
        waiting_time = 10
        while condition(self.seed_job):
            # Trying to avoid too much ssh polling by waiting if the loop time is less than a threshold
            loop_start_time = time.time()
            # Main loop
            print("ITERATION: %4i" % loops)
            self.print_status()
            # Sorting the queue to launch first the lower levels (bigger groups).
            self.sort(lower_first=True)
            # Launch one step (will loop through the jobs in the queue and take actions according to their status)
            change_counter += self.step()
            if change_counter > 10:
                self.checkpoint()
                change_counter = 0
            loop_end_time = time.time()
            if wait:
                while loop_end_time - loop_start_time < waiting_time:
                    time.sleep(1)
                    loop_end_time = time.time()
        print("done")

    def checkpoint(self, filename=None):
        from AutoRPE.UtilsWorkflow.CheckpointTools import checkpoint
        if filename is None:
            filename = self.checkpoint_name
        checkpoint(queue=self, iteration=self.loop, filename=filename)

    def __hash__(self):
        return self.seed_job.hash